import 'package:collection/collection.dart';

/// [10大排序算法](https://mp.weixin.qq.com/s?__biz=MzUyNjQxNjYyMg==&mid=2247484184&idx=1&sn=62965b401aa42107b3c17d1d8ea17454&chksm=fa0e6c99cd79e58f298e9026f677f912bd8c8e55edb48fc509b2b5834f05e529a9b47d59d202&mpshare=1&scene=24&srcid=06055KLg1Ouhw8jVE0VDPJ22&key=094361b14e5741f4bc7c219b7fadeceeff6e66e4c41c99e7cc74896924a5f6193f856acdbc2c6ebcc8fe7f0b26e8c6b9c2976bbec49d7830977e51554f5357cfabb0dd6a30e981e580b09867c25eb4fd&ascene=0&uin=MTMxMTQ0MTcyMQ%3D%3D&devicetype=iMac+MacBookPro11%2C2+OSX+OSX+10.15.2+build(19C57)&version=12020810&nettype=WIFI&lang=zh_CN&fontScale=100&exportkey=AfPIK3brrnEMX423B1tsDlU%3D&pass_ticket=hyZahRUDMJr4eMP4HEdGChvtp7d%2BDmtUpQz0%2BLwD4SZF38lFbnSfL37BBicSiQBl)
void main(List<String> args) {
  var data = [10, 8, 1, 11, 8, 20, 9, 10, 0];
  print(data.sortWithSelection(copy: true));
  print(data.sortWithQuick(copy: false));
}

// ignore: sdk_version_extension_methods
extension Sort on List {
  List sortWith(List Function(List source) sort, {bool copy = false}) {
    var source = copy ? List.from(this) : this;
    if (source.length < 2) {
      return source;
    }
    return sort(source);
  }

  /// 选择排序
  /// 原址排序，数组前半部分为已排序，后半部分未排序
  /// O(n^2) 不稳定
  List sortWithSelection({bool copy = false}) {
    return this.sortWith((source) {
      int length = source.length;
      for (int i = 0; i < length - 1; i++) {
        int min = i;
        for (int j = i + 1; j < length; j++) {
          if (source[j] < source[min]) {
            min = j;
          }
        }
        if (i != min) {
          var t = source[i];
          source[i] = source[min];
          source[min] = t;
        }
      }
      return source;
    }, copy: copy);
  }

  /// 快速排序
  /// 原址排序，分治处理，没次递归是，根据基准值划分子数组；基准索引pivot简单起见可谓左边界或有边界
  /// O(nlgn) 不稳定
  List sortWithQuick({bool copy = false}) {
    return this.sortWith((source) {
      //_quick(source, 0, source.length - 1);
      return _quick2(source);
    }, copy: copy);
  }

  void _swap(List array, int left, int right) {
    var t = array[left];
    array[left] = array[right];
    array[right] = t;
  }

  int _partition(List array, int left, int right) {
    var pivot = left;
    var index = pivot + 1;
    // 根据基准分区
    for (int i = index; i <= right; i++) {
      if (array[i] < array[pivot]) {
        _swap(array, i, index);
        index++;
      }
    }
    // 交换基准元素
    _swap(array, pivot, index - 1);
    return index - 1;
  }

  _quick(List array, int left, int right) {
    // 递归
    if (left < right) {
      var q = _partition(array, left, right);
      _quick(array, left, q - 1);
      _quick(array, q + 1, right);
    }
  }

  /// 非原址排序，思路与quick一致，利用了系统api，有隐藏开销
  List _quick2(List array) {
    if (array.length < 2) {
      return array;
    }
    var pivot = array[0];
    var less = array.sublist(1).where((e) => e <= pivot).toList();
    var greater = array.sublist(1).where((e) => e > pivot).toList();
    return _quick2(less)
      ..add(pivot)
      ..addAll(_quick2(greater));
  }
}
